@@ -1,6 +1,7 @@
 // Copyright (C) 2022 The Qt Company Ltd.
 // Copyright (C) 2019 Alexey Edelev <semlanik@gmail.com>, Viktor Kopp <vifactor@gmail.com>
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
+// Qt-Security score:critical reason:network-protocol
 
 #include <QtGrpc/private/qtgrpclogging_p.h>
 #include <QtGrpc/qgrpccalloptions.h>
@@ -67,6 +68,8 @@ using namespace QtGrpc;
     {serializationFormat}, or other options by constructing it with a
     QGrpcChannelOptions containing the required customizations.
 
+    \note \l{QGrpcChannelOptions::filterServerMetadata} is enabled by default.
+
     \section2 Transportation scheme
 
     The QGrpcHttp2Channel implementation prefers different transportation
@@ -98,6 +101,14 @@ using namespace QtGrpc;
         \li ✗
         \li QLocalSocket support \b{AND} scheme
         \li \c{unix:///tmp/grpc.socket}
+    \row
+        \li \c{unix-abstract}
+        \li Unix domain socket in abstract namespace
+        \li ✗
+        \li QLocalSocket support \b{AND}
+            \l{QLocalSocket::AbstractNamespaceOption}{AbstractNamespace}
+            support \b{AND} scheme
+        \li \c{unix-abstract:app_grpc_channel}
     \endtable
 
     \section2 Content-Type
@@ -156,6 +167,7 @@ Q_STATIC_LOGGING_CATEGORY(lcChannel, "qt.grpc.channel.http2")
 Q_STATIC_LOGGING_CATEGORY(lcStream, "qt.grpc.channel.http2.stream")
 
 constexpr QLatin1String UnixScheme("unix");
+constexpr QLatin1String UnixAbstractScheme("unix-abstract");
 constexpr QLatin1String HttpScheme("http");
 constexpr QLatin1String HttpsScheme("https");
 
@@ -313,6 +325,7 @@ public:
 
 private:
     [[nodiscard]] HPack::HttpHeader constructInitialHeaders() const;
+    [[nodiscard]] bool constructFilterServerMetadata() const;
     [[nodiscard]] QGrpcHttp2ChannelPrivate *channelPriv() const;
     [[nodiscard]] QGrpcHttp2Channel *channel() const;
     [[nodiscard]] bool handleContextExpired();
@@ -325,6 +338,7 @@ private:
     State m_state = State::Idle;
     const bool m_endStreamAtFirstData;
     bool m_writesDoneSent = false;
+    bool m_filterServerMetadata;
     QTimer m_deadlineTimer;
 
     Q_DISABLE_COPY_MOVE(Http2Handler)
@@ -334,7 +348,7 @@ class QGrpcHttp2ChannelPrivate : public QObject
 {
     Q_OBJECT
 public:
-    enum class SocketType : uint8_t { Tcp, Tls, Local };
+    enum class SocketType : uint8_t { Tcp, Tls, Local, LocalAbstract };
 
     explicit QGrpcHttp2ChannelPrivate(const QUrl &uri, QGrpcHttp2Channel *q);
     ~QGrpcHttp2ChannelPrivate() override = default;
@@ -409,7 +423,7 @@ private:
 Http2Handler::Http2Handler(QGrpcHttp2ChannelPrivate *parent, QGrpcOperationContext *context,
                            bool endStream)
     : QObject(parent), m_context(context), m_initialHeaders(constructInitialHeaders()),
-      m_endStreamAtFirstData(endStream)
+      m_endStreamAtFirstData(endStream), m_filterServerMetadata(constructFilterServerMetadata())
 {
     // If the context (lifetime bound to the user) is destroyed, this handler
     // can no longer perform any meaningful work. We allow it to be deleted;
@@ -536,6 +550,10 @@ HPack::HttpHeader Http2Handler::constructInitialHeaders() const
     const static QByteArray TEValue("trailers");
     const static QByteArray GrpcServiceNameHeader("service-name");
     const static QByteArray GrpcAcceptEncodingValue("identity,deflate,gzip");
+    const static QByteArray UserAgentHeader("user-agent");
+    const static QByteArray UserAgentValue("grpc-c++-qtgrpc/"_ba + QT_VERSION_STR + " ("_ba
+                                    + QSysInfo::productType().toUtf8() + '/'
+                                    + QSysInfo::productVersion().toUtf8() + ')');
 
     const auto &channelOptions = channel()->channelOptions();
     const auto *channel = channelPriv();
@@ -550,6 +568,7 @@ HPack::HttpHeader Http2Handler::constructInitialHeaders() const
         { ContentTypeHeader,        channel->contentType                     },
         { GrpcServiceNameHeader,    service                                  },
         { GrpcAcceptEncodingHeader, GrpcAcceptEncodingValue                  },
+        { UserAgentHeader,          UserAgentValue                           },
         { TEHeader,                 TEValue                                  },
     };
 
@@ -564,12 +583,19 @@ HPack::HttpHeader Http2Handler::constructInitialHeaders() const
         }
     };
 
-    iterateMetadata(channelOptions.metadata());
-    iterateMetadata(m_context->callOptions().metadata());
+    iterateMetadata(channelOptions.metadata(QtGrpc::MultiValue));
+    iterateMetadata(m_context->callOptions().metadata(QtGrpc::MultiValue));
 
     return headers;
 }
 
+bool Http2Handler::constructFilterServerMetadata() const
+{
+    return m_context->callOptions()
+        .filterServerMetadata()
+        .value_or(channel()->channelOptions().filterServerMetadata().value_or(true));
+}
+
 QGrpcHttp2ChannelPrivate *Http2Handler::channelPriv() const
 {
     return qobject_cast<QGrpcHttp2ChannelPrivate *>(this->parent());
@@ -750,7 +776,7 @@ void Http2Handler::handleHeaders(const HPack::HttpHeader &headers, HeaderPhase p
         false,
     };
 
-    QHash<QByteArray, QByteArray> metadata;
+    QMultiHash<QByteArray, QByteArray> metadata;
     std::optional<QtGrpc::StatusCode> statusCode;
     QString statusMessage;
 
@@ -796,9 +822,13 @@ void Http2Handler::handleHeaders(const HPack::HttpHeader &headers, HeaderPhase p
                       "[%p] Received unexcpected gRPC-reserved header: { key: %s, value: %s } "
                       "in phase: %s",
                       this, k.data(), v.data(), QDebug::toBytes(phase).data());
+        } else { // Custom-Metadata
+            metadata.insert(k, v);
+            continue;
         }
 
-        metadata.insert(k, v);
+        if (!m_filterServerMetadata)
+            metadata.insert(k, v);
     }
 
     if (validation.requireHttpStatus && !validation.hasHttpStatus) {
@@ -818,16 +848,14 @@ void Http2Handler::handleHeaders(const HPack::HttpHeader &headers, HeaderPhase p
 
     switch (phase) {
     case HeaderPhase::Initial:
-        m_context->setServerMetadata(std::move(metadata));
+        m_context->setServerInitialMetadata(std::move(metadata));
         break;
     case HeaderPhase::TrailersOnly:
         [[fallthrough]];
-    case HeaderPhase::Trailers: {
-        auto md = m_context->serverMetadata();
-        md.insert(metadata);
-        m_context->setServerMetadata(std::move(md));
+    case HeaderPhase::Trailers:
+        m_context->setServerTrailingMetadata(std::move(metadata));
         finish({ *statusCode, statusMessage });
-    } break;
+        break;
     default:
         Q_UNREACHABLE();
     }
@@ -893,9 +921,12 @@ QGrpcHttp2ChannelPrivate::QGrpcHttp2ChannelPrivate(const QUrl &uri, QGrpcHttp2Ch
         break;
     }
 
-    case SocketType::Local: {
+    case SocketType::Local:
+    case SocketType::LocalAbstract: {
 #if QT_CONFIG(localserver)
         auto socket = std::make_unique<QLocalSocket>();
+        if (socketType == SocketType::LocalAbstract)
+            socket->setSocketOptions(QLocalSocket::AbstractNamespaceOption);
         connect(socket.get(), &QLocalSocket::connected, this,
                 &QGrpcHttp2ChannelPrivate::createHttp2Connection);
         connect(socket.get(), &QLocalSocket::errorOccurred, this,
@@ -909,7 +940,7 @@ QGrpcHttp2ChannelPrivate::QGrpcHttp2ChannelPrivate(const QUrl &uri, QGrpcHttp2Ch
 #else
         m_reconnectFunction = [this] {
             qCFatal(lcChannel,
-                    "[%p] QLocalSocket support needed for 'unix' transportation",
+                    "[%p] QLocalSocket support needed for 'unix' or 'unix-abstract' transportation",
                     this);
         };
 #endif
@@ -1029,12 +1060,12 @@ QUrl QGrpcHttp2ChannelPrivate::sanitizeHostUri(const QUrl &rawUri,
         if (rawUri.scheme() != expected) {
             qCWarning(lcChannel,
                       "[%p] Unsupported transport protocol scheme '%s'. Fall back to '%s'.", this,
-                      qPrintable(rawUri.scheme()), qPrintable(expected));
+                      qPrintable(hostUri.scheme()), qPrintable(expected));
             sanitizedUri.setScheme(expected);
         }
     };
     const auto scheme = rawUri.scheme();
-    if (scheme == UnixScheme) {
+    if (scheme == UnixScheme || scheme == UnixAbstractScheme) {
         sanitizedUri.setScheme(HttpScheme);
     } else if (scheme == HttpsScheme || hasSslConfiguration(chOpts)) {
         check(HttpsScheme);
@@ -1054,6 +1085,8 @@ QGrpcHttp2ChannelPrivate::constructSocketType(const QUrl &rawUri, const QGrpcCha
     const auto scheme = rawUri.scheme();
     if (scheme == UnixScheme)
         return SocketType::Local;
+    if (scheme == UnixAbstractScheme)
+        return SocketType::LocalAbstract;
     if (scheme == HttpsScheme || hasSslConfiguration(chOpts))
         return SocketType::Tls;
     return SocketType::Tcp;
@@ -1071,8 +1104,8 @@ QByteArray QGrpcHttp2ChannelPrivate::setupContentTypeNegotiation(QGrpcHttp2Chann
     bool warnAboutFormatConflict = !formatSuffix.isEmpty();
     QByteArray finalContentType = contentTypeFromOptions;
 
-    const auto it = channelOptions.metadata().constFind(ContentTypeHeader.data());
-    if (it != channelOptions.metadata().cend()) {
+    const auto it = channelOptions.metadata(QtGrpc::MultiValue).constFind(ContentTypeHeader.data());
+    if (it != channelOptions.metadata(QtGrpc::MultiValue).cend()) {
         if (formatSuffix.isEmpty() && it.value() != DefaultContentType) {
             // Auto-detect format from content-type header
             if (it.value() == "application/grpc+json") {
